import os
from collections.abc import Generator

import pytest

from core.model_runtime.entities.llm_entities import LLMResult, LLMResultChunk, LLMResultChunkDelta
from core.model_runtime.entities.message_entities import (
    AssistantPromptMessage,
    ImagePromptMessageContent,
    PromptMessageTool,
    SystemPromptMessage,
    TextPromptMessageContent,
    UserPromptMessage,
)
from core.model_runtime.entities.model_entities import AIModelEntity, ModelType
from core.model_runtime.errors.validate import CredentialsValidateFailedError
from core.model_runtime.model_providers.openai.llm.llm import OpenAILargeLanguageModel

"""FOR MOCK FIXTURES, DO NOT REMOVE"""
from tests.integration_tests.model_runtime.__mock.openai import setup_openai_mock


def test_predefined_models():
    model = OpenAILargeLanguageModel()
    model_schemas = model.predefined_models()

    assert len(model_schemas) >= 1
    assert isinstance(model_schemas[0], AIModelEntity)


@pytest.mark.parametrize("setup_openai_mock", [["chat"]], indirect=True)
def test_validate_credentials_for_chat_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    with pytest.raises(CredentialsValidateFailedError):
        model.validate_credentials(model="gpt-3.5-turbo", credentials={"openai_api_key": "invalid_key"})

    model.validate_credentials(model="gpt-3.5-turbo", credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")})


@pytest.mark.parametrize("setup_openai_mock", [["completion"]], indirect=True)
def test_validate_credentials_for_completion_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    with pytest.raises(CredentialsValidateFailedError):
        model.validate_credentials(model="text-davinci-003", credentials={"openai_api_key": "invalid_key"})

    model.validate_credentials(
        model="text-davinci-003", credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")}
    )


@pytest.mark.parametrize("setup_openai_mock", [["completion"]], indirect=True)
def test_invoke_completion_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-3.5-turbo-instruct",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY"), "openai_api_base": "https://api.openai.com"},
        prompt_messages=[UserPromptMessage(content="Hello World!")],
        model_parameters={"temperature": 0.0, "max_tokens": 1},
        stream=False,
        user="abc-123",
    )

    assert isinstance(result, LLMResult)
    assert len(result.message.content) > 0
    assert model._num_tokens_from_string("gpt-3.5-turbo-instruct", result.message.content) == 1


@pytest.mark.parametrize("setup_openai_mock", [["completion"]], indirect=True)
def test_invoke_stream_completion_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-3.5-turbo-instruct",
        credentials={
            "openai_api_key": os.environ.get("OPENAI_API_KEY"),
            "openai_organization": os.environ.get("OPENAI_ORGANIZATION"),
        },
        prompt_messages=[UserPromptMessage(content="Hello World!")],
        model_parameters={"temperature": 0.0, "max_tokens": 100},
        stream=True,
        user="abc-123",
    )

    assert isinstance(result, Generator)

    for chunk in result:
        assert isinstance(chunk, LLMResultChunk)
        assert isinstance(chunk.delta, LLMResultChunkDelta)
        assert isinstance(chunk.delta.message, AssistantPromptMessage)
        assert len(chunk.delta.message.content) > 0 if chunk.delta.finish_reason is None else True


@pytest.mark.parametrize("setup_openai_mock", [["chat"]], indirect=True)
def test_invoke_chat_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-3.5-turbo",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(content="Hello World!"),
        ],
        model_parameters={
            "temperature": 0.0,
            "top_p": 1.0,
            "presence_penalty": 0.0,
            "frequency_penalty": 0.0,
            "max_tokens": 10,
        },
        stop=["How"],
        stream=False,
        user="abc-123",
    )

    assert isinstance(result, LLMResult)
    assert len(result.message.content) > 0


@pytest.mark.parametrize("setup_openai_mock", [["chat"]], indirect=True)
def test_invoke_chat_model_with_vision(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-4-vision-preview",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(
                content=[
                    TextPromptMessageContent(
                        data="Hello World!",
                    ),
                    ImagePromptMessageContent(
                        mime_type="image/png",
                        format="png",
                        base64_data=" ",
                    ),
                ]
            ),
        ],
        model_parameters={"temperature": 0.0, "max_tokens": 100},
        stream=False,
        user="abc-123",
    )

    assert isinstance(result, LLMResult)
    assert len(result.message.content) > 0


@pytest.mark.parametrize("setup_openai_mock", [["chat"]], indirect=True)
def test_invoke_chat_model_with_tools(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-3.5-turbo",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(
                content="what's the weather today in London?",
            ),
        ],
        model_parameters={"temperature": 0.0, "max_tokens": 100},
        tools=[
            PromptMessageTool(
                name="get_weather",
                description="Determine weather in my location",
                parameters={
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state e.g. San Francisco, CA"},
                        "unit": {"type": "string", "enum": ["c", "f"]},
                    },
                    "required": ["location"],
                },
            ),
            PromptMessageTool(
                name="get_stock_price",
                description="Get the current stock price",
                parameters={
                    "type": "object",
                    "properties": {"symbol": {"type": "string", "description": "The stock symbol"}},
                    "required": ["symbol"],
                },
            ),
        ],
        stream=False,
        user="abc-123",
    )

    assert isinstance(result, LLMResult)
    assert isinstance(result.message, AssistantPromptMessage)
    assert len(result.message.tool_calls) > 0


@pytest.mark.parametrize("setup_openai_mock", [["chat"]], indirect=True)
def test_invoke_stream_chat_model(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    result = model.invoke(
        model="gpt-3.5-turbo",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(content="Hello World!"),
        ],
        model_parameters={"temperature": 0.0, "max_tokens": 100},
        stream=True,
        user="abc-123",
    )

    assert isinstance(result, Generator)

    for chunk in result:
        assert isinstance(chunk, LLMResultChunk)
        assert isinstance(chunk.delta, LLMResultChunkDelta)
        assert isinstance(chunk.delta.message, AssistantPromptMessage)
        assert len(chunk.delta.message.content) > 0 if chunk.delta.finish_reason is None else True
        if chunk.delta.finish_reason is not None:
            assert chunk.delta.usage is not None
            assert chunk.delta.usage.completion_tokens > 0


def test_get_num_tokens():
    model = OpenAILargeLanguageModel()

    num_tokens = model.get_num_tokens(
        model="gpt-3.5-turbo-instruct",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[UserPromptMessage(content="Hello World!")],
    )

    assert num_tokens == 3

    num_tokens = model.get_num_tokens(
        model="gpt-3.5-turbo",
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(content="Hello World!"),
        ],
        tools=[
            PromptMessageTool(
                name="get_weather",
                description="Determine weather in my location",
                parameters={
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state e.g. San Francisco, CA"},
                        "unit": {"type": "string", "enum": ["c", "f"]},
                    },
                    "required": ["location"],
                },
            ),
        ],
    )

    assert num_tokens == 72


@pytest.mark.parametrize("setup_openai_mock", [["chat", "remote"]], indirect=True)
def test_fine_tuned_models(setup_openai_mock):
    model = OpenAILargeLanguageModel()

    remote_models = model.remote_models(credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")})

    if not remote_models:
        assert isinstance(remote_models, list)
    else:
        assert isinstance(remote_models[0], AIModelEntity)

    for llm_model in remote_models:
        if llm_model.model_type == ModelType.LLM:
            break

    assert isinstance(llm_model, AIModelEntity)

    # test invoke
    result = model.invoke(
        model=llm_model.model,
        credentials={"openai_api_key": os.environ.get("OPENAI_API_KEY")},
        prompt_messages=[
            SystemPromptMessage(
                content="You are a helpful AI assistant.",
            ),
            UserPromptMessage(content="Hello World!"),
        ],
        model_parameters={"temperature": 0.0, "max_tokens": 100},
        stream=False,
        user="abc-123",
    )

    assert isinstance(result, LLMResult)


def test__get_num_tokens_by_gpt2():
    model = OpenAILargeLanguageModel()
    num_tokens = model._get_num_tokens_by_gpt2("Hello World!")

    assert num_tokens == 3
